/*
 * $Id: Communication.cs 289 2007-06-12 09:27:37Z lloydd $
 */

using System;
using System.Text;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.Collections;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using Mono.Unix;

namespace WiMo.RAPI
{
   public class Communication : CommunicationBase
   {
      public Communication (Connection connection) : base (connection)
      {
      }
      
      public void Process ()
      {
         Exception ee = null;
         
         Connection.LockStream ();

         LastError = 0xffffffff;
         LastResult = 0xffffffff;
         
         try {
            Send ();
            Receive ();
         }
         catch (Exception e) {
            ee = e;
         }
         
         Connection.UnlockStream ();
         
         if (ee != null)
            throw new CommunicationException (this, "Process message failed", ee);
      }
      
      private void Send ()
      {
         if (write_buffer.Length < 1)
            throw new CommunicationException (this, "WriteBuffer is empty");
         
         try {
            byte[] buf = new byte[write_buffer.Length + 4];
            BitConverter.GetBytes (write_buffer.Length).CopyTo (buf, 0);
            write_buffer.CopyTo (buf, 4);
            
            if (Debug)
               PrintBuffer ("send", buf);
            
            Stream.Write (buf, 0, buf.Length);
         }
         catch (Exception e) {
            throw new CommunicationException (this, "Send message failed", e);
         }
         
         ClearWrite ();
      }
      
      public virtual void Receive ()
      {
         int size = 0;
         
         ClearRead ();
         read_buffer = new byte [4];

         try {
            // packet size
            if (Stream.Read (read_buffer, 0, sizeof (uint)) != sizeof (uint))
               throw new CommunicationException (this, "Failed to read response size");
               
            size = ReadInt32 ();
            
            ClearRead ();
            
            read_buffer = new byte [size];
            int readed = 0;
            
            // packet
            while (readed != size)
               readed += Stream.Read (read_buffer, readed, size - readed);
            
            if (Debug)
               PrintReadBuffer ();

            if (ReadUInt32 () == 1) {
               ReadLastError ();
               ClearRead ();
               return;
            }
         }
         catch (Exception e) {
            throw new CommunicationException (this, "Receive response failed", e);
         }
      }
      
      protected void ClearRead ()
      {
         read_index = 0;
         read_buffer = new byte [0];
      }

      protected void ClearWrite ()
      {
         write_index = 0;
         write_buffer = new byte [0];
      }
      
      private object Read (Type o_type, int o_size)
      {
         object o = null;

         if ((read_index + o_size) > read_buffer.Length)
            throw new CommunicationException (this, 
               string.Format ("ReadBuffer(length:{0}, index:{1}) is too small " +
                  "for requested type of size {2}",
                  read_buffer.Length, read_index, o_size));
         
         if (o_type == typeof(byte))
            o = (byte) read_buffer [read_index];
         
         else if (o_type == typeof(byte[])) {
            byte[] b = new byte [o_size];
            for (int i = 0; i < o_size; i++)
               b [i] = read_buffer [read_index + i];
            o = b;
            
         } else if (o_type == typeof(double))
            o = (double) BitConverter.ToDouble (read_buffer, read_index);
            
         else if (o_type == typeof(uint))
            o = (uint) BitConverter.ToUInt32 (read_buffer, read_index);
         
         else if (o_type == typeof(int))
            o = (int) BitConverter.ToInt32 (read_buffer, read_index);
         
         else if (o_type == typeof(ushort))
            o = (ushort) BitConverter.ToUInt16 (read_buffer, read_index);
         
         else if (o_type == typeof(short))
            o = (short) BitConverter.ToInt16 (read_buffer, read_index);
         
         else if (o_type == typeof(string)) {
            if (o_size == 0) {
               int size = BitConverter.ToInt32 (read_buffer, read_index);
               read_index += sizeof (int);
               o_size = (size + 1) * 2;
               
               if (size > 0) {
                  if ((read_index + o_size) > read_buffer.Length)
                     throw new CommunicationException (this, 
                        string.Format ("ReadBuffer(length:{0}, index:{1}) is too small " +
                           "for requested type of size {2}",
                           read_buffer.Length, read_index, o_size));
                  o = Encoding.Unicode.GetString(read_buffer, read_index, size * 2 );
               } else
                  o = "";
            } else
               o = Encoding.Unicode.GetString(read_buffer, read_index, o_size);
            
            string s = (string) o;
            
            if (s.Length > 0)
               if (s [s.Length - 1] == '\0')
                  s = s.Remove (s.Length - 1, 1);
            
            o = s;
         }
        
         if (o == null)
            throw new CommunicationException (this, 
               string.Format ("No value for type {0}", o_type));
         
         read_index += o_size;
         
         return o;
      }

      private void Write (object o)
      {
         byte[] buf = null;
         
         if (o is byte)
            buf = new byte [1] { (byte) o };
         else if (o is byte[])
            buf = (byte[]) o;
         else if (o is Double)
            buf = BitConverter.GetBytes ((double) o);
         else if (o is UInt32)
            buf = BitConverter.GetBytes ((UInt32) o);
         else if (o is Int32)
            buf = BitConverter.GetBytes ((Int32) o);
         else if (o is UInt16)
            buf = BitConverter.GetBytes ((UInt16) o);
         else if (o is Int16)
            buf = BitConverter.GetBytes ((Int16) o);
         else if (o is string) {
            WriteUInt32 (1);
            WriteUInt32 ((uint) (((string) o).Length + 1));
            WriteByteArray (Encoding.Unicode.GetBytes((string) o));
            WriteByteArray (new byte [2] {0, 0});
            return;
         } else if (o is Filetime) {
            Filetime f = (Filetime) o;
            WriteUInt32 (f.Low);
            WriteUInt32 (f.High);
            return;
         }
            
         if (buf == null)
            throw new CommunicationException (this, 
               string.Format ("No value to write for type {0}", o.GetType ()));
         
         if (write_buffer.Length < (write_index + buf.Length)) {
            byte[] ba = new byte [write_index + buf.Length];
            write_buffer.CopyTo (ba, 0);
            write_buffer = ba;
         }
         
         buf.CopyTo (write_buffer, write_index);
         write_index += buf.Length;
      }

      protected void ReadLastError ()
      {
         LastError = ReadUInt32 ();
         
         if (Debug)
            Console.WriteLine ("LastError: 0x{0}", LastError.ToString ("X8"));
      }
      
      protected void ReadResult ()
      {
         LastResult = ReadUInt32 ();

         if (Debug)
            Console.WriteLine ("LastResult: 0x{0}", LastResult.ToString ("X8"));
      }
      
      protected void ReadOptional (out uint data_size)
      {
         uint have_parameter = 0;
         uint have_value = 0;
         
         data_size = 0;
         
         // have parameter
         have_parameter = ReadUInt32 ();
         
         if (Debug)
            Console.WriteLine ("HaveParameter: {0}", have_parameter.ToString ("X8"));
         
         if (have_parameter != 1)
            return;
         
         // size
         data_size = ReadUInt32 ();

         // have value
         if ((have_value = ReadUInt32 ()) != 1)
            data_size = 0;
         
         if (Debug) {
            Console.WriteLine ("DataSize: {0}", data_size);
            Console.WriteLine ("HaveValue: {0}", have_value);
         }
      }
      
      protected void ReadOptional (uint required_size)
      {
         uint size = 0;
         
         ReadOptional (out size);
         
         if (size != required_size)
            throw new CommunicationException (this, string.Format (
               "Readed optional value size {} doesn't match required size {1}",
               size, required_size));
      }
      
      protected byte ReadByte ()
      {
         return (byte) Read (typeof (byte), sizeof (byte));
      }

      protected ushort ReadUInt16 ()
      {
         return (ushort) Read (typeof (ushort), sizeof (ushort));
      }

      protected short ReadInt16 ()
      {
         return (short) Read (typeof (short), sizeof (short));
      }

      protected uint ReadUInt32 ()
      {                                                  
         return (uint) Read (typeof (uint), sizeof (uint));
      }
      
      protected int ReadInt32 ()
      {
         return (int) Read (typeof (int), sizeof (int));
      }
      
      protected double ReadDouble ()
      {
         return (double) Read (typeof (double), sizeof (double));
      }
      
      protected byte[] ReadByteArray (int len)
      {
         return (byte[]) Read (typeof (byte[]), len);
      }

      protected byte[] ReadByteArrayFrom (int pos, int len)
      {
         int orig_index = ReadIndex;
         ReadIndex = pos;
         byte[] b = (byte[]) Read (typeof (byte[]), len);
         ReadIndex = orig_index;
         return b;
      }

      protected Filetime ReadFiletime ()
      {
         uint low = ReadUInt32 ();
         uint high = ReadUInt32 ();
         return new Filetime (high, low);
      }
      
      protected string ReadString ()
      {
         return ReadString (0);
      }
      
      protected string ReadString (int length)
      {
         return (string) Read (typeof(string), length);
      }
      
      protected string ReadStringFrom (int pos)
      {
         int orig_index = ReadIndex;
         string str = "";
         string s = "";
         
         ReadIndex = pos;
         
         while ((s = ReadString (2)).Length > 0)
            str += s;
         
         ReadIndex = orig_index;
         
         return str;
      }
      
      public uint ReadOptionalUInt32 ()
      {
         ReadOptional (sizeof(uint));
         return (uint) Read (typeof(uint), sizeof(uint));
      }

      protected void WriteOptional (uint value_size, bool send_data)
      {
         WriteUInt32 (1);
         WriteUInt32 (value_size);
         WriteInt32  (send_data ? 1 : 0);
      }
      
      protected void WriteByte (byte v)
      {
         Write (v);
      }

      protected void WriteUInt16 (ushort v)
      {
         Write (v);
      }

      protected void WriteInt16 (short v)
      {
         Write (v);
      }

      protected void WriteUInt32 (uint v)
      {                                                  
         Write (v);
      }
      
      protected void WriteInt32 (int v)
      {
         Write (v);
      }
      
      protected void WriteDouble (double v)
      {
         Write (v);
      }
      
      protected void WriteFiletime (Filetime v)
      {
         Write (v);
      }
      
      protected void WriteString (string v)
      {
         Write (v);
      }

      protected void WriteStringTo (string v, int pos)
      {
         int orig_index = WriteIndex;
         WriteIndex = pos;
         WriteString (v);
         WriteIndex = orig_index;
      }
      
      protected void WriteRawString (string v)
      {
         WriteByteArray (Encoding.Unicode.GetBytes (v));
         WriteByteArray (new byte [2] {0, 0});
      }
      protected void WriteRawStringTo (string v, int pos)
      {
         int orig_index = WriteIndex;
         WriteIndex = pos;
         WriteRawString (v);
         WriteIndex = orig_index;
      }
      
      protected void WriteByteArray (byte[] v)
      {
         Write (v);
      }

      protected void WriteByteArrayTo (byte[] v, int pos)
      {
         int orig_index = WriteIndex;
         WriteIndex = pos;
         WriteByteArray (v);
         WriteIndex = orig_index;
      }
      
      protected void WriteOptionalUInt32 (uint i, bool send_data)
      {
         WriteOptional (sizeof (uint), send_data);
         if (send_data)
            WriteUInt32 (i);
      }
      
      protected void WriteOptionalString (string str, bool send_data)
      {
         WriteOptional ((uint) ((str.Length + 1) * 2), send_data);
         if (send_data)
            WriteRawString (str);
      }
   }
}
