/*
 * $Id: Connection.cs 289 2007-06-12 09:27:37Z lloydd $
 */

using System;
using System.Text;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.Collections;
using System.Collections.Generic;

namespace WiMo.RAPI
{
   public class Connection : IDisposable
   {
      private bool disposed = false;

      private Mutex stream_mutex;
      
      private string rapi_server;
      private int rapi_port;
      
      private uint key;
      private string password;
      
      private TcpClient client;
      
      private List<Context> contexts;
      private List<Context2> contexts2;
      
      private bool debug;
      
      public Connection (string rapi_server) :
         this (rapi_server, 990)
      {
      }

      public Connection (string rapi_server, string password, uint key) :
         this (rapi_server, 990, password, key)
      {
      }
      
      public Connection (string rapi_server, int rapi_port) :
         this (rapi_server, rapi_port, "", 0)
      {
      }

      public Connection (string rapi_server, int rapi_port, string password, uint key)
      {
         this.stream_mutex = new Mutex ();
         
         this.rapi_server = rapi_server;
         this.rapi_port = rapi_port;
         this.password = password;
         this.key = key;
         this.contexts = new List<Context> ();
         this.contexts2 = new List<Context2> ();
         this.debug = false;
         
         Connect ();
      }
      
      ~Connection ()
      {
         Dispose (false);
      }
      
      public bool Debug {
         get {
            return debug;
         }
         set {
            debug = value;
         }
      }
      
      public int ReceiveTimeout {
         get {
            return client.ReceiveTimeout;
         }
         set {
            client.ReceiveTimeout = value;
         }
      }
      
      public int SendTimeout {
         get {
            return client.SendTimeout;
         }
         set {
            client.SendTimeout = value;
         }
      }
      
      public string Server {
         get {
            return rapi_server;
         }
         set {
            rapi_server = value;
         }
      }
      
      public int Port {
         get {
            return rapi_port;
         }
         set {
            rapi_port = value;
         }
      }
      
      public string Password {
         get {
            return password;
         }
         set {
            password = value;
         }
      }

      public uint Key {
         get {
            return key;
         }
         set {
            key = value;
         }
      }
      
      public bool IsEncrypted {
         get {
            return Password.Length > 0;
         }
      }
      
      public NetworkStream Stream {
         get {
            return client.GetStream ();
         }
      }

      public Context[] Contexts {
         get {
            return contexts.ToArray ();
         }
      }
      
      public Context2[] Contexts2 {
         get {
            return contexts2.ToArray ();
         }
      }
      
      public Context CreateContext ()
      {
         Context ctx = null;
         
         lock (contexts) {
            
            ctx = new Context (this);
            
            contexts.Add (ctx);
         }
         
         return ctx;
      }
      
      public Context2 CreateContext2 ()
      {
         Context2 ctx = null;
         
         lock (contexts2) {
            
            ctx = new Context2 (this);
            
            contexts2.Add (ctx);
         }
         
         return ctx;
      }
      
      public void DestroyContext (Context ctx)
      {
         lock (contexts) {
            contexts.Remove (ctx);
         }
      }

      public void DestroyContext2 (Context2 ctx)
      {
         lock (contexts2) {
            contexts2.Remove (ctx);
         }
      }

      public void LockStream ()
      {
         stream_mutex.WaitOne ();
      }

      public void UnlockStream ()
      {
         stream_mutex.ReleaseMutex ();
      }
      
      public void Dispose ()
      {
         Dispose (true);
         
         GC.SuppressFinalize (this);
      }
      
      private void Connect ()
      {
         try {
            
            client = new TcpClient ();
            
            client.Client.SetSocketOption (SocketOptionLevel.Socket, SocketOptionName.ReceiveTimeout, 4000);
            client.Client.SetSocketOption (SocketOptionLevel.Socket, SocketOptionName.SendTimeout, 4000);
            
            client.Connect (Server, Port);
            
            client.Client.SetSocketOption (SocketOptionLevel.Socket, SocketOptionName.KeepAlive, 15000);
            
            ReceiveTimeout = 8000;
            SendTimeout = 8000;

            if (!client.Connected)
               throw new Exception ("Connect failed");
            
            SendPassword ();
         }
         catch (Exception e) {
            
            throw new Exception (string.Format (
               "TCP connect to RAPI server {0}:{1} failed: {2}", Server, Port, e.Message));
         }
      }
      
      private void SendPassword ()
      {
         if (!IsEncrypted)
            return;
         
         if (!SendPassword (Password, Key))
            throw new Exception ("Invalid password");
      }

      private bool SendPassword (string psw, uint key)
      {
         byte[] psw_unicode = Encoding.Unicode.GetBytes (psw);
         byte[] encoded = new byte [psw_unicode.Length + 2];
         
         psw_unicode.CopyTo (encoded, 0);

         for (int i = 0; i < encoded.Length; i++)
            encoded [i] = (byte)(encoded [i] ^ key);
         
         byte[] len = BitConverter.GetBytes ((Int16) encoded.Length);
         
         NetworkStream stream = client.GetStream ();
         stream.Write (len, 0, len.Length);
         stream.Write (encoded, 0, encoded.Length);
         
         if (stream.Read (len, 0, 1) != 1)
            return false;
         
         return len [0] == 1;
      }
      
      private void Dispose (bool disposing)
      {
         if (this.disposed)
            return;

         if (disposing)
            // disconnect socket to rapi server
            if (client != null) {
               
               client.Close ();
               
               client = null;
            }
         
         this.disposed = true;
      }
   }
}
