/*
 * $Id: Context.cs 289 2007-06-12 09:27:37Z lloydd $
 */

using System;
using System.Text;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.Collections;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using Mono.Unix;

namespace WiMo.RAPI
{
   public class Context : Communication, ISystem, IFilesystem, IDatabase
   {
      private int CEDB_MAXDBASENAMELEN = 32;
      private int CEDB_MAX_SORT_ORDER = 4;
      private int CEPROPVAL_SIZE = sizeof (uint) + 2 * sizeof (short) + 8;
      
      protected int rapi_version;
      
      public Context (Connection connection) : base (connection)
      {
         rapi_version = 1;
      }

      public int RapiVersion {
         get {
            return rapi_version;
         }
      }

      public int MaxDbaseNameLength {
         get {
            return CEDB_MAXDBASENAMELEN;
         }
      }
      
      public int MaxSortOrder {
         get {
            return CEDB_MAX_SORT_ORDER;
         }
      }
      
      // FILESYSTEM
      public bool CeCreateDirectory (string path)
      {
         ClearWrite ();
         
         WriteUInt32          (0x17);
         WriteOptionalString  (path, true);
         WriteUInt32          (0); //SecurityAttributes
         
         Process ();
         ReadLastError ();
         ReadResult ();
         
         return LastResult == 1;
      }

      public bool CeRemoveDirectory (string path)
      {
         ClearWrite ();
         
         WriteUInt32          (0x18);
         WriteOptionalString  (path, true);
         
         Process ();
         ReadLastError ();
         ReadResult ();
         
         return LastResult == 1;
      }

      public string CeGetSpecialFolderPath (SpecialFolder folder)
      {
         ClearWrite ();
         
         WriteUInt32 (0x44);
         WriteUInt32 ((uint) folder);
         WriteUInt32 (512);
         
         Process ();
         ReadLastError ();

         return ReadString ();
      }

      // "Zkontrolovat jestli neni prehozene ShareMode a FileShareMode
      public FileHandle CeCreateFile (string filename,
                                ShareMode desired_access,
                                FileShareMode share_mode,
                                CreationDisposition creation_disposition,
                                FileAttribute flags_and_attributes,
                                FileHandle template_file)
      {
         ClearWrite ();
         
         WriteUInt32 (0x05);
         WriteUInt32 ((uint) desired_access);
         WriteUInt32 ((uint) share_mode);
         WriteUInt32 ((uint) creation_disposition);
         WriteUInt32 ((uint) flags_and_attributes);
         WriteUInt32 (template_file.Value);
         WriteString (filename);
         
         Process ();
         ReadLastError ();
         ReadResult ();

         if (LastResult == 0 || LastResult == 0xffffffff)
            return FileHandle.Zero;
         
         return new FileHandle (LastResult);
      }

      public bool CeCloseHandle (IValueHandle handle)
      {
         ClearWrite ();
         
         WriteUInt32 (0x08);
         WriteUInt32 (handle.Value);
         
         Process ();
         ReadLastError ();
         ReadResult ();
         
         return LastResult == 1;
      }

      public bool CeDeleteFile (string filename)
      {
         ClearWrite ();
         
         WriteUInt32          (0x1c);
         WriteOptionalString  (filename, true);
         
         Process ();
         ReadLastError ();
         ReadResult ();
         
         return LastResult == 1;
      }

      public bool CeCopyFile (string existing_filename,
                              string new_filename,
                              bool fail_if_exists)
      {
         ClearWrite ();
         
         WriteUInt32          (0x1b);
         WriteOptionalString  (existing_filename, true);
         WriteOptionalString  (new_filename, true);
         WriteUInt32          ((uint) (fail_if_exists ? 1 : 0));
         
         Process ();
         ReadLastError ();
         ReadResult ();
         
         return LastResult == 1;
      }

      public bool CeMoveFile (string existing_filename,
                              string new_filename)
      {
         ClearWrite ();
         
         WriteUInt32          (0x1a);
         WriteOptionalString  (existing_filename, true);
         WriteOptionalString  (new_filename, true);
         
         Process ();
         ReadLastError ();
         ReadResult ();
         
         return LastResult == 1;
      }

      public bool CeGetFileSize (FileHandle handle,
                                 out uint file_size_high,
                                 out uint file_size_low)
      {
         ClearWrite ();
         
         file_size_high = 0;
         file_size_low = 0;
         
         WriteUInt32 (0x1d);
         WriteUInt32 (handle.Value);
         WriteUInt32 (1);
         WriteUInt32 (sizeof (uint));
         WriteUInt32 (0);
         
         Process ();
         ReadLastError ();
         ReadResult ();

         if (LastResult != 1)
            return false;
         
         // doplnit kontrolu vysledku operace
         
         file_size_high = ReadOptionalUInt32 ();
         file_size_low = (uint) LastResult;
         
         return true;
      }
      
      public uint CeWriteFile (FileHandle handle, byte[] buffer) // bytes_written
      {
         ClearWrite ();
         
         WriteUInt32    (0x07);
         WriteUInt32    (handle.Value);
         WriteUInt32    (1);
         WriteUInt32    ((uint)buffer.Length);
         WriteByteArray (buffer);
         WriteUInt32    (0);
         
         Process ();
         ReadLastError ();
         ReadResult ();

         if (LastResult != 1)
            return 0;
         
         return ReadUInt32 ();
      }
      
      public uint CeReadFile (FileHandle handle, out byte[] buffer, uint bytes_to_read) // bytes_readed
      {
         ClearWrite ();
         
         buffer = new byte [0];
         
         WriteUInt32 (0x06);
         WriteUInt32 (handle.Value);
         WriteUInt32 (1);
         WriteUInt32 (bytes_to_read);
         WriteUInt32 (0);
         WriteUInt32 (0);
         
         Process ();
         ReadLastError ();
         ReadResult ();

         if (LastResult != 1)
            return 0;

         uint bytes_readed = ReadUInt32 ();
         
         if (bytes_readed > 0)
            buffer = ReadByteArray ((int) bytes_readed);
         
         return (uint) buffer.Length;
      }

      public FindData[] CeFindAllFiles (string path, FindFileAttribute flags)
      {
         ClearWrite ();
         
         WriteUInt32 (0x09);
         WriteString (path);
         WriteUInt32 ((uint) flags);
         
         Process ();

         uint count = ReadUInt32 ();
         
         FindData[] fnd = new FindData [count];
         
         uint name_size = 0;
         
         if (count > 0)
            for (int idx = 0; idx < count; idx++) {
               if ((flags & FindFileAttribute.Name) != 0)
                  name_size = ReadUInt32 ();
   
               if ((flags & FindFileAttribute.Attributes) != 0)
                  fnd [idx].FileAttributes = (FileAttribute) (ReadUInt32 ());
   
               if ((flags & FindFileAttribute.Creation_Time) != 0)
                  fnd [idx].CreationTime = ReadFiletime ();

               if ((flags & FindFileAttribute.Last_Access_Time) != 0)
                  fnd [idx].LastAccessTime = ReadFiletime ();

               if ((flags & FindFileAttribute.Last_Write_Time) != 0)
                  fnd [idx].LastWriteTime = ReadFiletime ();

               if ((flags & FindFileAttribute.Size_High) != 0)
                  fnd [idx].FileSizeHigh = ReadUInt32 ();

               if ((flags & FindFileAttribute.Size_Low) != 0)
                  fnd [idx].FileSizeLow = ReadUInt32 ();
               
               if ((flags & FindFileAttribute.Oid) != 0)
                  fnd [idx].Oid = new Oid (ReadUInt32 ());

               if ((flags & FindFileAttribute.Name) != 0)
                  fnd [idx].FileName = ReadString ((int) name_size * 2);
            }
         
         return fnd;
      }

      private void ReadFindData (ref FindData find_data)
      {
         int size = ReadInt32 ();
         
         find_data.FileAttributes = (FileAttribute) ReadUInt32 ();
         find_data.CreationTime = ReadFiletime ();
         find_data.LastAccessTime = ReadFiletime ();
         find_data.LastWriteTime = ReadFiletime ();
         find_data.FileSizeHigh = ReadUInt32 ();
         find_data.FileSizeLow = ReadUInt32 ();
         find_data.Oid = new Oid (ReadUInt32 ());

         size -= 10 * sizeof (uint);
         find_data.FileName = ReadString (size);
      }
      
      public FileFindHandle CeFindFirstFile (string filename,
                                             out FindData find_data)
      {
         ClearWrite ();
         
         find_data = new FindData ();
         
         WriteUInt32 (0x00);
         WriteString (filename);
         
         Process ();
         ReadLastError ();
         ReadResult ();

         if (LastResult == 0 || LastResult == 0xffffffff)
            return FileFindHandle.Zero;

         ReadFindData (ref find_data);
         
         return new FileFindHandle (LastResult);
      }

      public FileHandle CeFindNextFile (FileFindHandle file_find_handle,
                                        out FindData find_data)
      {
         ClearWrite ();
         
         find_data = new FindData ();
         
         WriteUInt32 (0x01);
         WriteUInt32 (file_find_handle.Value);
         
         Process ();
         ReadLastError ();
         ReadResult ();

         if (LastResult == 0 || LastResult == 0xffffffff)
            return FileHandle.Zero;

         ReadFindData (ref find_data);
         
         return new FileHandle (LastResult);
      }
      
      public bool CeFindClose (FileFindHandle file_find_handle)
      {
         ClearWrite ();
         
         WriteUInt32 (0x02);
         WriteUInt32 (file_find_handle.Value);
         
         Process ();
         ReadLastError ();
         ReadResult ();
         
         return LastResult == 1;
      }
      
      public bool CeGetFileAttributes (string filename, out FileAttribute file_attr)
      {
         ClearWrite ();
         
         file_attr = (FileAttribute) 0;
         
         WriteUInt32 (0x03);
         WriteString (filename);
         
         Process ();
         ReadLastError ();
         ReadResult ();
         
         if (LastResult == 0xffffffff)
            return false;
         
         file_attr = (FileAttribute) LastResult;
         
         return true;
      }
      
      public bool CeSetFileAttributes (string filename, FileAttribute file_attr)
      {
         ClearWrite ();
         
         WriteUInt32 (0x04);
         WriteUInt32 ((uint) file_attr);
         WriteString (filename);
         
         Process ();
         ReadLastError ();
         ReadResult ();
         
         return LastResult == 1;
      }
      
      public bool CeSHCreateShortcut (string shortcut, string target)
      {
         ClearWrite ();
         
         WriteUInt32          (0x30);
         WriteOptionalString  (shortcut, true);
         WriteOptionalString  (target, true);
         
         Process ();
         ReadLastError ();
         ReadResult ();
         
         return LastResult == 1;
      }
      
      // DATABASE
      public DbaseFindData[] CeFindAllDatabases (uint dbase_type,
                                                 FindDatabaseAttribute flags)
      {
         ClearWrite ();
         
         WriteUInt32 (0x2c);
         WriteUInt32 (dbase_type);
         WriteUInt16 ((ushort) flags);
         
         Process ();
         
         ushort count = ReadUInt16 ();
         
         DbaseFindData[] fnd = new DbaseFindData [count];
         
         int name_size = 0;
         
         if (count > 0)
            for (int idx = 0; idx < count; idx++) {
               fnd [idx].DbInfo = new DbaseInfo ();
               
               if ((flags & FindDatabaseAttribute.Oid) != 0)
                  fnd [idx].Oid = new Oid (ReadUInt32 ());
   
               if ((flags & FindDatabaseAttribute.Name) != 0)
                  name_size = ReadInt32 ();
   
               if ((flags & FindDatabaseAttribute.Flags) != 0)
                  fnd [idx].DbInfo.Flags = (FindDatabaseAttribute) (ReadUInt32 ());

               if ((flags & FindDatabaseAttribute.Name) != 0)
                  fnd [idx].DbInfo.DbaseName = ReadString (name_size * 2);
   
               if ((flags & FindDatabaseAttribute.Type) != 0)
                  fnd [idx].DbInfo.DbaseType = ReadUInt32 ();

               if ((flags & FindDatabaseAttribute.Num_Records) != 0)
                  fnd [idx].DbInfo.NumRecords = ReadUInt16 ();
               
               if ((flags & FindDatabaseAttribute.Num_Sort_Order) != 0)
                  fnd [idx].DbInfo.NumSortOrder = ReadUInt16 ();
   
               if ((flags & FindDatabaseAttribute.Size) != 0)
                  fnd [idx].DbInfo.Size = ReadUInt32 ();

               if ((flags & FindDatabaseAttribute.Last_Modified) != 0)
                  fnd [idx].DbInfo.LastModified = ReadFiletime ();

               if ((flags & FindDatabaseAttribute.Sort_Spec) != 0) {
                  fnd [idx].DbInfo.SortSpecs =
                     new SortOrderSpec [MaxSortOrder];
                  
                  for (int j = 0; j < MaxSortOrder; j++ ) {
                     fnd [idx].DbInfo.SortSpecs [j].PropId =
                        new PropId (ReadUInt32 ());
                     fnd [idx].DbInfo.SortSpecs [j].Flags =
                        (FindDatabaseAttribute) (ReadUInt32 ());
                  }
               } else
                  fnd [idx].DbInfo.SortSpecs = new SortOrderSpec [0]; 
            }
         
         return fnd;
      }

      public DbFindHandle CeFindFirstDatabase (uint dbase_type)
      {
         ClearWrite ();
         
         WriteUInt32 (0x0a);
         WriteUInt32 (dbase_type);
         
         Process ();
         ReadLastError ();
         ReadResult ();

         if (LastResult == 0 || LastResult == 0xffffffff)
            return DbFindHandle.Zero;
         
         return new DbFindHandle (LastResult);
      }

      public Oid CeFindNextDatabase (DbFindHandle find_handle)
      {
         ClearWrite ();
         
         WriteUInt32 (0x0b);
         WriteUInt32 (find_handle.Value);
         
         Process ();
         ReadLastError ();
         ReadResult ();

         if (LastResult == 0 || LastResult == 0xffffffff)
            return Oid.Zero;
         
         return new Oid (LastResult);
      }

      public Oid CeCreateDatabase (string dbase_name, uint dbase_type,
                                    SortOrderSpec[] sort_specs)
      {
         ClearWrite ();
         
         WriteUInt32 (0x0d);
         WriteUInt32 (dbase_type);
         WriteUInt16 ((ushort) (sort_specs == null ? 0 : sort_specs.Length));
         
         if (sort_specs != null)
            for (int i = 0; i < sort_specs.Length; i++) {
               WriteUInt32 (sort_specs [i].PropId.Value);
               WriteUInt32 ((uint) sort_specs [i].Flags);
            }
         
         WriteString (dbase_name);
            
         Process ();
         ReadLastError ();
         ReadResult ();

         if (LastResult == 0 || LastResult == 0xffffffff)
            return Oid.Zero;
         
         return new Oid (LastResult);
      }

      public bool CeDeleteDatabase (Oid oid)
      {
         ClearWrite ();
         
         WriteUInt32 (0x0f);
         WriteUInt32 (oid.Value);
            
         Process ();
         ReadLastError ();
         ReadResult ();
         
         return LastResult == 1;
      }
      
      public DbHandle CeOpenDatabase (Oid oid,
                                      string dbase_name,
                                      PropId prop_id,
                                      FindDatabaseAttribute flags)
      {
         ClearWrite ();
         
         Oid open_oid = new Oid (oid.Value);
         
         if (open_oid == Oid.Zero) {
            if (dbase_name.Length == 0)
               throw new Exception ("Oid and database name not set");
            else {
               // find database by name
               foreach (DbaseFindData d in CeFindAllDatabases (0, flags))
                  if (d.DbInfo.DbaseName == dbase_name) {
                     open_oid = d.Oid;
                     break;
                  }
                  
               if (open_oid == Oid.Zero)
                  return DbHandle.Zero;
            }
         }
         
         WriteUInt32 (0x0e);
         WriteUInt32 (open_oid.Value);
         WriteUInt32 (prop_id.Value);
         WriteUInt32 ((uint) flags);
         
         Process ();
         ReadLastError ();
         ReadResult ();

         if (LastResult == 0 || LastResult == 0xffffffff)
            return DbHandle.Zero;
         
         return new DbHandle (LastResult);
      }

      public Oid CeReadRecordProps (DbHandle handle,
                                    FindDatabaseAttribute flags,
                                    out PropVal[] prop_val_array)
      {
         ClearWrite ();
         
         prop_val_array = new PropVal [0];
         
         WriteUInt32 (0x10);
         WriteUInt32 (handle.Value);
         WriteUInt32 ((uint) flags);
         WriteUInt32 (0);
         WriteUInt32 (0);
         WriteUInt32 (0);
         WriteUInt32 (0);
            
         Process ();
         ReadLastError ();
         ReadResult ();

         if (LastResult == 0 || LastResult == 0xffffffff)
            return Oid.Zero;
         
         ReadInt32 (); // size
         ushort count = ReadUInt16 ();
         
         prop_val_array = new PropVal [count];
         
         int begin_pos = ReadIndex;
         
         for (int i = 0; i < count; i++) {
            PropVal p = new PropVal ();
            
            p.PropId = new PropId (ReadUInt32 ());
            
            ReadUInt16 (); // data_len_unused
            
            p.Flags = ReadUInt16 ();
            
            switch (p.PropId.ValType) {
               case PropValType.Int16:
                  p.Int16 = ReadInt16 ();
                  ReadIndex += (8 - sizeof (short));
                  break;
                  
               case PropValType.Int32:
                  p.Int32 = ReadInt32 ();
                  ReadIndex += (8 - sizeof (int));
                  break;
                  
               case PropValType.Double:
                  p.Double = ReadDouble ();
                  break;
                  
               case PropValType.Bool:
                  p.Bool = ReadUInt32 () != 0;
                  ReadIndex += (8 - sizeof (uint));
                  break;
                  
               case PropValType.UInt16:
                  p.UInt16 = ReadUInt16 ();
                  ReadIndex += (8 - sizeof (ushort));
                  break;
                  
               case PropValType.UInt32:
                  p.UInt32 = ReadUInt32 ();
                  ReadIndex += (8 - sizeof (uint));
                  break;
                  
               case PropValType.Filetime:
                  p.Filetime = ReadFiletime ();
                  break;
                  
               case PropValType.String:
                  int pos = ReadInt32 ();
                  ReadIndex += (8 - sizeof (int));
                  p.String = ReadStringFrom (begin_pos + pos);
                  break;
                  
               case PropValType.Blob:
                  int len = ReadInt32 ();
                  pos = ReadInt32 ();
                  p.Blob = ReadByteArrayFrom (begin_pos + pos, len);
                  break;
            }
            prop_val_array [i] = p;
         }
         
         return new Oid (LastResult);
      }

      public bool CeWriteRecordProps (DbHandle handle,
                                     ref Oid oid,
                                     PropVal[] props)
      {
         ClearWrite ();
         
         WriteUInt32 (0x11);
         WriteUInt32 (handle.Value);
         WriteUInt32 (oid.Value);
         WriteUInt16 ((ushort) props.Length);
         
         int total_size = props.Length * CEPROPVAL_SIZE;
         int array_size = total_size;
         
         foreach (PropVal p in props) {
            switch (p.PropId.ValType) {
               case PropValType.String:
                  total_size += 2 * (p.String.Length + 1);
                  break;
               case PropValType.Blob:
                  total_size += p.Blob.Length;
                  break;
            }
         }
         
         WriteUInt32 ((uint) total_size);
         
         int begin_pos = WriteIndex;
                                       
         foreach (PropVal p in props) {
            WriteUInt32 (p.PropId.Value);
            WriteUInt16 (0);
            WriteUInt16 (p.Flags);
            
            switch (p.PropId.ValType) {
               case PropValType.Filetime:
                  WriteFiletime (p.Filetime);
                  break;
               case PropValType.Double:
                  WriteDouble (p.Double);
                  break;
               case PropValType.Bool:
                  WriteInt32 (p.Bool ? 1 : 0);
                  WriteByteArray (new byte [4]);
                  break;
               case PropValType.Int32:
                  WriteInt32 (p.Int32);
                  WriteByteArray (new byte [4]);
                  break;
               case PropValType.UInt32:
                  WriteUInt32 (p.UInt32);
                  WriteByteArray (new byte [4]);
                  break;
               case PropValType.Int16:
                  WriteInt16 (p.Int16);
                  WriteByteArray (new byte [6]);
                  break;
               case PropValType.UInt16:
                  WriteUInt16 (p.UInt16);
                  WriteByteArray (new byte [6]);
                  break;
               case PropValType.String:
                  WriteInt32 (array_size);
                  WriteByteArray (new byte [4]);
                  WriteRawStringTo (p.String, begin_pos + array_size);
                  array_size += 2 * (p.String.Length + 1);
                  break;
               case PropValType.Blob:
                  WriteInt32 (p.Blob.Length);
                  WriteInt32 (array_size);
                  WriteByteArrayTo (p.Blob, begin_pos + array_size);
                  array_size += p.Blob.Length;
                  break;
               default:
                  break;
            }
         }
         
         Process ();
         ReadLastError ();
         ReadResult ();

         if (LastResult == 0 || LastResult == 0xffffffff)
            return false;

         oid = new Oid (LastResult);
         
         return true;
      }

      public Oid CeSeekDatabase (DbHandle handle, DatabaseSeekType seek_type,
                                  uint value, out uint index)
      {
         ClearWrite ();
         
         index = 0xffffffff;
         
         WriteUInt32 (0x13);
         WriteUInt32 (handle.Value);
         WriteUInt32 ((uint) seek_type);

         switch (seek_type) {
            case DatabaseSeekType.ValueSmaller:
            case DatabaseSeekType.ValueFirstEqual:
            case DatabaseSeekType.ValueGreater:
            case DatabaseSeekType.ValueNextEqual:
               throw new NotImplementedException ("Seek by value is not yet implemented");
            default:
               WriteUInt32 (value);
               break;
         }
         
         Process ();
         ReadLastError ();
         ReadResult ();

         if (LastResult == 0 || LastResult == 0xffffffff)
            return Oid.Zero;
         
         index = ReadUInt32 ();
         
         return new Oid (LastResult);
      }

      public bool CeDeleteRecord (DbHandle handle, Oid oid)
      {
         ClearWrite ();
         
         WriteUInt32 (0x12);
         WriteUInt32 (handle.Value);
         WriteUInt32 (oid.Value);
         
         Process ();
         ReadLastError ();
         ReadResult ();
         
         return LastResult == 1;
      }
      
      public bool CeSetDatabaseInfo (Oid oid, DbaseInfo new_info)
      {
         ClearWrite ();
         
         WriteUInt32 (0x14);
         WriteUInt32 (oid.Value);
         
         string name = new_info.DbaseName.Substring (0,
                        Math.Min (new_info.DbaseName.Length,MaxDbaseNameLength));
         
         WriteUInt32    ((uint) new_info.Flags);
         WriteByteArray (Encoding.Unicode.GetBytes (name));
         
         if (name.Length < MaxDbaseNameLength) {
            int size = 2 * (MaxDbaseNameLength - name.Length);
            WriteByteArray (new byte [size]);
         }
         
         WriteUInt32    (new_info.DbaseType);
         WriteUInt32    ((uint) new_info.NumRecords);
         WriteUInt32    ((uint) new_info.NumSortOrder);
         WriteUInt32    (new_info.Size);
         WriteFiletime  (new_info.LastModified);

         for (int j = 0; j < MaxSortOrder; j++ )
            if (j < new_info.SortSpecs.Length) {
               WriteUInt32 (new_info.SortSpecs [j].PropId.Value);
               WriteUInt32 ((uint) new_info.SortSpecs [j].Flags);
            } else {
               WriteUInt32 (0);
               WriteUInt32 (0);
            }
         
         Process ();
         ReadLastError ();
         ReadResult ();
         
         return LastResult == 1;
      }
      
      // SYSTEM
      public bool CeGetStoreInformation (out StoreInformation si)
      {
         uint requested_size = 2 * sizeof (uint);
         
         si = new StoreInformation ();
         
         ClearWrite ();
         
         WriteUInt32    (0x29);
         WriteOptional  (requested_size, false);
         
         Process ();
         ReadLastError ();
         ReadResult ();
         
         if (LastResult != 1)
            return false;
         
         ReadOptional (requested_size);

         si.StoreSize = ReadUInt32 ();
         si.FreeSize = ReadUInt32 ();
         
         return true;
      }
      
      public SystemInfo CeGetSystemInfo ()
      {
         uint requested_size = 7 * sizeof (uint) + 4 * sizeof (ushort);
         
         ClearWrite ();
         
         WriteUInt32    (0x2f);
         WriteOptional  (requested_size, false);
         
         Process ();
         ReadLastError ();
         
         ReadOptional (requested_size);

         SystemInfo si = new SystemInfo ();
         si.ProcessorArch = (ProcessorArch) ReadUInt16 ();
         si.Reserved = ReadUInt16 ();
         si.PageSize = ReadUInt32 ();
         si.MinimumApplicationAddress = ReadUInt32 ();
         si.MaximumApplicationAddress = ReadUInt32 ();
         si.ActiveProcessorMask = ReadUInt32 ();
         si.NumberOfProcessors = ReadUInt32 ();
         si.ProcessorType = (ProcessorType) ReadUInt32 ();
         si.AllocationGranularity = ReadUInt32 ();
         si.ProcessorLevel = ReadUInt16 ();
         si.ProcessorRevision = ReadUInt16 ();
         
         return si;
      }
      
      public bool CeGetSystemPowerStatusEx (out SystemPowerStatusEx ps, bool refresh)
      {
         uint requested_size = 8 * sizeof (byte) + 4 * sizeof (uint);
         
         ps = new SystemPowerStatusEx ();
         
         ClearWrite ();
         
         WriteUInt32    (0x41);
         WriteOptional  (requested_size, false);
         WriteInt32     (refresh ? 1 : 0);
         
         Process ();
         ReadLastError ();
         ReadResult ();
         
         if (LastResult != 1)
            return false;
         
         ReadOptional (requested_size);
         
         ps.AcLineStatus = (ACStatus) ReadByte ();
         ps.BatteryFlag = (BatteryFlag) ReadByte ();
         ps.BatteryLifePercent = ReadByte ();
         ps.Reserved1 = ReadByte ();
         ps.BatteryLifeTime = ReadUInt32 ();
         ps.BatteryFullLifeTime = ReadUInt32 ();
         ps.Reserved2 = ReadByte ();
         ps.BackupBatteryFlag = (BatteryFlag) ReadByte ();
         ps.BackupBatteryLifePercent = ReadByte ();
         ps.Reserved3 = ReadByte ();
         ps.BackupBatteryLifeTime = ReadUInt32 ();
         ps.BackupBatteryFullLifeTime = ReadUInt32 ();
         
         return true;
      }

      public bool CeGetVersionEx (out VersionInfo vi)
      {
         ClearWrite ();
         
         vi = new VersionInfo ();
         
         WriteUInt32 (0x3b);
         
         Process ();
         ReadLastError ();
         ReadResult ();

         if (LastResult != 1)
            return false;
         
         ReadUInt32 (); // size
         ReadUInt32 (); // OsVersionInfoSize
         
         vi.MajorVersion = ReadUInt32 ();
         vi.MinorVersion = ReadUInt32 ();
         vi.BuildNumber = ReadUInt32 ();
         vi.Platform = (PlatformType) ReadUInt32 ();
         
         return true;
      }

      public bool CeGetSystemMemoryDivision (out uint storage_pages,
                                             out uint ram_pages,
                                             out uint page_size)
      {
         ClearWrite ();
         
         storage_pages = 0;
         ram_pages = 0;
         page_size = 0;
         
         WriteUInt32          (0x28);
         WriteOptionalUInt32  (0, false);
         WriteOptionalUInt32  (0, false);
         WriteOptionalUInt32  (0, false);
         
         Process ();
         ReadLastError ();
         ReadResult ();

         if (LastResult != 1)
            return false;

         storage_pages = ReadOptionalUInt32 ();
         ram_pages = ReadOptionalUInt32 ();
         page_size = ReadOptionalUInt32 ();
         
         return true;
      }

      public bool CeSetSystemMemoryDivision (uint storage_pages)
      {
         ClearWrite ();
         
         WriteUInt32 (0x42);
         WriteUInt32 (storage_pages);
         
         Process ();
         ReadLastError ();
         ReadResult ();
         
         return LastResult == 0;
      }
      
      public bool CeSyncTimeToPc ()
      {
         ClearWrite ();
         
         WriteUInt32    (0x37);
         WriteFiletime  (Filetime.Now);
         WriteUInt32    (0);
         WriteUInt32    (10000);
         
         Process ();
         ReadLastError ();
         
         return true;
      }

      public bool CeRegCopyFile (string filename)
      {
         ClearWrite ();
         
         WriteUInt32          (0x2d);
         WriteOptionalString  (filename, true);
         
         Process ();
         ReadLastError ();
         ReadResult ();
         
         return LastResult == 1;
      }
      
      public bool CeRegRestoreFile (string filename)
      {
         ClearWrite ();
         
         WriteUInt32          (0x2e);
         WriteOptionalString  (filename, true);
         
         Process ();
         ReadLastError ();
         ReadResult ();
         
         return LastResult == 1;
      }
      
      public bool CeKillAllApps ()
      {
         ClearWrite ();
         
         WriteUInt32 (0x49);
         
         Process ();
         ReadLastError ();
         ReadResult ();
         
         return LastResult == 1;
      }

      public bool CeCreateProcess (string application_name,
                                   string command_line,
                                   out ProcessInformation process_info)
      {
         uint request_size = 4 * sizeof (uint);

         process_info = new ProcessInformation ();

         ClearWrite ();
         
         WriteUInt32          (0x19);
         WriteOptionalString  (application_name, true);
         WriteOptionalString  (command_line, true);
         WriteUInt32          (0);
         WriteUInt32          (0);
         WriteUInt32          (0);
         WriteUInt32          (0);
         WriteUInt32          (0);
         WriteUInt32          (0);
         WriteUInt32          (0);
         WriteUInt32          (1);
         WriteUInt32          (request_size);
         WriteUInt32          (0);
         
         Process ();
         ReadLastError ();
         ReadResult ();
         
         if (LastResult != 1)
            return false;
         
         ReadOptional (request_size);
         
         process_info.Process = ReadUInt32 ();
         process_info.Thread = ReadUInt32 ();
         process_info.ProcessId = ReadUInt32 ();
         process_info.ThreadId = ReadUInt32 ();
         
         return true;
      }

      public bool CeCheckPassword (string password)
      {
         ClearWrite ();
         
         WriteUInt32          (0x34);
         WriteOptionalString  (password, true);
         
         Process ();
         ReadLastError ();
         ReadResult ();
         
         return LastResult == 1;
      }
   }
}
